import logging
from querent.logging.custom_formatter import CustomFormatter

def create_console_handler(log_format, date_format):
    console_handler = logging.StreamHandler()
    formatter = CustomFormatter(fmt=log_format, datefmt=date_format)
    console_handler.setFormatter(formatter)
    return console_handler


def create_file_handler(log_file, log_format, date_format):
    file_handler = logging.FileHandler(log_file)
    formatter = CustomFormatter(fmt=log_format, datefmt=date_format)
    file_handler.setFormatter(formatter)
    return file_handler



import os
import re
from email.header import decode_header

from bs4 import BeautifulSoup


class EmailReader:
    def clean_email_body(self, email_body):
        """
        Function to clean the email body.

        Args:
            email_body (str): The email body to be cleaned.

        Returns:
            str: The cleaned email body.
        """
        if email_body is None:
            email_body = ""
        email_body = BeautifulSoup(email_body, "html.parser")
        email_body = email_body.get_text()
        email_body = "".join(email_body.splitlines())
        email_body = " ".join(email_body.split())
        email_body = email_body.encode("ascii", "ignore")
        email_body = email_body.decode("utf-8", "ignore")
        email_body = re.sub(r"http\S+", "", email_body)
        return email_body

    def clean(self, text):
        """
        Function to clean the text.

        Args:
            text (str): The text to be cleaned.

        Returns:
            str: The cleaned text.
        """
        return "".join(c if c.isalnum() else "_" for c in text)

    def obtain_header(self, msg):
        """
        Function to obtain the header of the email.

        Args:
            msg (email.message.Message): The email message.

        Returns:
            str: The From field of the email.
        """
        if msg["Subject"] is not None:
            Subject, encoding = decode_header(msg["Subject"])[0]
        else:
            Subject = ""
            encoding = ""
        if isinstance(Subject, bytes):
            try:
                if encoding is not None:
                    Subject = Subject.decode(encoding)
                else:
                    Subject = ""
            except [LookupError] as err:
                pass
        From = msg["From"]
        To = msg["To"]
        Date = msg["Date"]
        return From, To, Date, Subject

    def download_attachment(self, part, subject):
        """
        Function to download the attachment from the email.

        Args:
            part (email.message.Message): The email message.
            subject (str): The subject of the email.

        Returns:
            None
        """
        filename = part.get_filename()
        if filename:
            folder_name = self.clean(subject)
            if not os.path.isdir(folder_name):
                os.mkdir(folder_name)
                filepath = os.path.join(folder_name, filename)
                open(filepath, "wb").write(part.get_payload(decode=True))



import json
import re
from unidecode import unidecode


import spacy
from querent.config.core.gpt_llm_config import GPTConfig
from querent.common.types.ingested_images import IngestedImages
from querent.kg.rel_helperfunctions.openai_functions import FunctionRegistry
from querent.common.types.querent_event import EventState, EventType
from querent.core.base_engine import BaseEngine
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.ingested_messages import IngestedMessages
from querent.common.types.ingested_code import IngestedCode
from querent.common.types.querent_queue import QuerentQueue
from querent.kg.rel_helperfunctions.embedding_store import EmbeddingStore
from querent.kg.rel_helperfunctions.triple_to_json import TripleToJsonConverter
from querent.logging.logger import setup_logger
from openai import OpenAI
from querent.common.types.file_buffer import FileBuffer
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
)
from dotenv import load_dotenv, find_dotenv
import json

_ = load_dotenv(find_dotenv())


class GPTNERLLM(BaseEngine):
    def __init__(
        self,
        input_queue: QuerentQueue,
        config: GPTConfig
    ):
        self.logger = setup_logger(__name__, "OPENAINERLLM")
        try:
            self.nlp = spacy.load(config.spacy_model_path)
            super().__init__(input_queue)
            self.file_buffer = FileBuffer()
            self.rel_model_name = config.rel_model_name
            if config.openai_api_key:
                self.gpt_llm = OpenAI(api_key=config.openai_api_key)
            else:
                self.gpt_llm = OpenAI()
            self.function_registry = FunctionRegistry()
            self.create_emb = EmbeddingStore(inference_api_key=config.huggingface_token)
            self.user_context = config.user_context
        
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to Initialize. {e}")
            raise Exception(f"Invalid {self.__class__.__name__} configuration. Unable to Initialize. {e}")
    
    def validate(self) -> bool:
        return True

    def process_messages(self, data: IngestedMessages):
        return super().process_messages(data)
    
    def process_images(self, data: IngestedImages):
        return super().process_messages(data)

    async def process_code(self, data: IngestedCode):
        return super().process_messages(data)

    @staticmethod
    def validate_ingested_tokens(data: IngestedTokens) -> bool:
        if data.is_error():
            
            return False

        return True
    
    def get_context(self, sentences):
        contexts = []
        for i, sentence in enumerate(sentences):
            # Previous sentence or empty string if current sentence is the first
            prev_sentence = sentences[i-1] if i > 0 else ''
            # Next sentence or empty string if current sentence is the last
            next_sentence = sentences[i+1] if i < len(sentences)-1 else ''
            # Current context
            context = f"{prev_sentence} {sentence} {next_sentence}".strip()
            contexts.append(context)
        return contexts
    
    @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6))
    def completion_with_backoff(self, **kwargs):
        return self.gpt_llm.chat.completions.create(**kwargs)
    
    def generate_response(self, messages):
        response = self.completion_with_backoff(
            model=self.rel_model_name,
            messages=messages,
            temperature=0,
        )
        return response

    def extract_semantic_triples(self, chat_completion):
    # Extract the message content from the ChatCompletion
        message_content = chat_completion.choices[0].message.content.replace('\n', '')

        # Parse the JSON content into a Python list
        try:
            triples_list = eval(message_content)
            if not isinstance(triples_list, list):
                raise ValueError("Content is not a list")
        except Exception as e:
            raise ValueError(f"Error parsing content: {e}")

        return triples_list

    def filter_relevant_triples(self, triples, context, max_distance):
        def tokenize_text(text):
            # Tokenize the text while ignoring common filler words
            filler_words = ["the", "and", "in", "of", "a", "an", "to", "with"]
            tokens = re.findall(r'\b\w+\b', text)
            tokens = [token.lower() for token in tokens if token.lower() not in filler_words]
            return tokens

        def calculate_distance(tokenized_text, subject, object):
            # Find the start index of subject and object in tokenized_text
            subject_tokens = tokenize_text(subject)
            object_tokens = tokenize_text(object)
            
            subject_start_index = None
            object_start_index = None
            
            for i in range(len(tokenized_text)):
                if tokenized_text[i:i+len(subject_tokens)] == subject_tokens:
                    subject_start_index = i
                    break
            
            for i in range(len(tokenized_text)):
                if tokenized_text[i:i+len(object_tokens)] == object_tokens:
                    object_start_index = i
                    break
            
            if subject_start_index is not None and object_start_index is not None:
                return abs(subject_start_index - object_start_index)
            else:
                return float('inf')  # Return a large distance if either subject or object is not found

        tokenized_context = tokenize_text(context.lower())  # Convert context to lowercase
        relevant_triples = []
        
        for triple in triples:
            subject = triple['subject'].lower()  # Convert subject to lowercase
            object = triple['object'].lower()    # Convert object to lowercase
            
            distance = calculate_distance(tokenized_context, subject, object)
            
            if distance <= max_distance:
                triple['subject'] = subject
                triple['object'] = object
                triple['subject_type'] = triple['subject_type'].lower().replace(" ", "_")
                triple['object_type'] = triple['object_type'].lower().replace(" ", "_")
                triple['predicate'] = triple['predicate'].lower()
                triple['predicate_type'] = triple['predicate_type'].lower().replace(" ", "_")
                triple['sentence'] = context  # Add the context as a key-value pair
                relevant_triples.append(triple)
        
        return relevant_triples
    
    def remove_duplicate_triplets(self,input_list):
        seen = set()
        result = []

        for item in input_list:
            triplet = (item['subject'], item['predicate'], item['object'])
            if triplet not in seen:
                result.append(item)
                seen.add(triplet)

        return result
      
    async def process_tokens(self, data: IngestedTokens):
        try:
            if not GPTNERLLM.validate_ingested_tokens(data):
                    self.set_termination_event()                    
                    return 

            if data.data:
                single_string = ' '.join(data.data)
                clean_text = unidecode(single_string)
            else:
                clean_text = data.data
            if not data.is_token_stream : 
                file, content = self.file_buffer.add_chunk(
                data.get_file_path(), clean_text)
            else:
                content = clean_text
                file = data.get_file_path()
            if content: 
                doc = self.nlp(content)
                sentences = [sent.text for sent in doc.sents]
                contexts = self.get_context(sentences)
                final_triples = []
                for context in contexts:
                    identify_entity_message = f"""Please analyze the provided context below. Once you have understood the context, answer the user query using the specified output format.
        
                    Context: {contexts[0]}

                    Output Format:
                    [
                        {{
                            'subject': 'Identified as the main entity in the context, typically the initiator or primary focus of the action or topic being discussed.',
                            'predicate': 'The relationship (predicate) between the subject and the object.',
                            'object': 'This parameter represents the entity in the context directly impacted by or involved in the action, typically the recipient or target of the main verb's action.',
                            'subject_type': 'The category of the subject entity e.g. location, person, event, material, process etc.',
                            'object_type': 'The category of the object entity e.g. location, person, event, material, process etc.',
                            'predicate_type': 'The category of the predicate e.g. causative, action, ownership, occurance etc.'
                        }},
                        {{}},  # Additional triples go here
                        {{}},  # Additional triples go here
                        # ...  # Additional triples go here
                    ]
                    """
                    if not self.user_context:
                        messages_classify_entity = [
                            {"role": "user", "content": identify_entity_message},
                            {"role": "user", "content": "Query: First, identify all geological entities in the provided context. Then, create relevant semantic triples (Subject, Predicate, Object) and also categorize the respective the Subject, Object types (e.g. location, person, event, material, process etc.) and Predicate type. Use the above output format to provide all the relevant semantic triples."},
                        ]
                    else :
                        messages_classify_entity = [
                            {"role": "user", "content": identify_entity_message},
                            {"role": "user", "content": self.user_context},
                        ]
                    identify_entity_response = self.generate_response(messages_classify_entity)
                    try:
                        semantic_triples = self.extract_semantic_triples(identify_entity_response)
                        relevant_triples = self.filter_relevant_triples(semantic_triples, context, 10)
                    except Exception as e:
                        self.logger.info(f"Error extracting semantic triples in GPT NER & LLM Class: {e}")
                        continue
                    if len(relevant_triples)>0:
                        final_triples.extend(relevant_triples)
                #final_triples = [{'subject': 'paleocene–eocene thermal maximum (petm) record', 'predicate': 'is present in', 'object': '543-m-thick (1780 ft) deep-marine section in the gulf of mexico (gom)', 'subject_type': 'event', 'object_type': 'location', 'predicate_type': 'occurrence', 'sentence': 'ABSTRACT In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM.'}, {'subject': 'climate and tectonic perturbations', 'predicate': 'can induce', 'object': 'a substantial response', 'subject_type': 'process', 'object_type': 'event', 'predicate_type': 'causative', 'sentence': 'ABSTRACT In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM.'}, {'subject': 'gulf coastal plain', 'predicate': 'is', 'object': 'ultimately in the gom', 'subject_type': 'location', 'object_type': 'location', 'predicate_type': 'ownership', 'sentence': 'ABSTRACT In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM.'}, {'subject': 'paleocene–eocene thermal maximum (petm) record', 'predicate': 'is present in', 'object': '543-m-thick (1780 ft) deep-marine section in the gulf of mexico (gom)', 'subject_type': 'event', 'object_type': 'location', 'predicate_type': 'occurrence', 'sentence': 'ABSTRACT In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.'}, {'subject': 'climate and tectonic perturbations', 'predicate': 'can induce', 'object': 'a substantial response', 'subject_type': 'process', 'object_type': 'event', 'predicate_type': 'causative', 'sentence': 'ABSTRACT In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.'}, {'subject': 'upstream north american catchments', 'predicate': 'can induce', 'object': 'a substantial response', 'subject_type': 'location', 'object_type': 'event', 'predicate_type': 'causative', 'sentence': 'ABSTRACT In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.'}, {'subject': 'downstream sectors of the gulf coastal plain', 'predicate': 'can be impacted by', 'object': 'upstream north american catchments', 'subject_type': 'location', 'object_type': 'location', 'predicate_type': 'causative', 'sentence': 'ABSTRACT In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.'}, {'subject': 'climate and tectonic perturbations', 'predicate': 'can induce', 'object': 'a substantial response', 'subject_type': 'process', 'object_type': 'event', 'predicate_type': 'causative', 'sentence': 'We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.'}, {'subject': 'upstream north american catchments', 'predicate': 'can induce', 'object': 'a substantial response', 'subject_type': 'location', 'object_type': 'event', 'predicate_type': 'causative', 'sentence': 'We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.'}]
                final_triples = self.remove_duplicate_triplets(final_triples)
                if len(final_triples) > 0:
                    for triple in final_triples:
                        graph_json = json.dumps(triple)
                        if graph_json:
                            current_state = EventState(EventType.Graph,1.0, graph_json, file)
                            await self.set_state(new_state=current_state)
                        context_embeddings = self.create_emb.get_embeddings([triple['sentence']])[0]
                        triple['context_embeddings'] = context_embeddings
                        triple['context'] = triple['sentence']
                        vector_json = json.dumps(TripleToJsonConverter.convert_vectorjson((triple['subject'],json.dumps(triple), triple['object'])))
                        if vector_json:
                                current_state = EventState(EventType.Vector,1.0, vector_json, file)
                                await self.set_state(new_state=current_state)

        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to extract predicates using GPT NER LLM class. {e}")
            raise Exception(f"An error occurred while extracting predicates using GPT NER LLM class: {e}")
    
    async def process_messages(self, data: IngestedMessages):
        raise NotImplementedError



import re
from typing import List

"""
    FixedEntityExtractor is a class designed to extract specific entities and their related sentences from a given text. It uses spaCy for natural language processing and regular expressions for pattern matching.

    Key functionalities include:
    - Initialization with a list of fixed entities, entity types, and a spaCy model.
    - Creation of a combined regular expression pattern to match fixed entities in text.
    - Lemmatization of provided entities for better matching accuracy.
    - Extraction of sentences containing the specified entities from the provided text.
    - Ability to include contextual sentences adjacent to those containing the entities.
    - Calculation of text reduction percentage when filtering sentences.
    - Processing and filtering entity pairs based on specified entity types and fixed entities.

    This class is particularly useful for scenarios where only certain types of entities or specific named entities are of interest in a larger text.
    """

class FixedEntityExtractor:
    def __init__(self, fixed_entities: List[str] = None, entity_types: List[str] = None, model=None):
        try:
            self.nlp = model
        except Exception as e:
            raise Exception(f"Error loading spaCy model: {e}")

        self.entity_types = entity_types
        self.fixed_entities = fixed_entities

        if fixed_entities:
            try:
                self.entity_pattern = self.create_combined_pattern(fixed_entities)
                self.lemmatized_entities = set(self.lemmatize_entities(fixed_entities))
            except Exception as e:
                raise Exception(f"Error in processing fixed entities: {e}")

    def create_combined_pattern(self, entities):
        try:
            combined_pattern = '|'.join(map(re.escape, entities))
            return re.compile(r'\b(?:' + combined_pattern + r')\b', re.IGNORECASE)
        except Exception as e:
            raise Exception(f"Error creating combined pattern: {e}")

    def lemmatize_entities(self, entities):
        try:
            return [self.nlp(entity)[0].lemma_ for entity in entities]
        except Exception as e:
            raise Exception(f"Error in lemmatizing entities: {e}")

    def find_entity_sentences(self, text: str, chunk_size=1000) -> str:
        try:
            doc = self.nlp(text)
        except Exception as e:
            raise Exception(f"Error processing text with spaCy: {e}")

        relevant_sentences = []
        added_sentences = set()
        prev_sentence = None
        for j, sentence in enumerate(doc.sents):
            try:
                if self.is_entity_present(sentence):
                    self.add_contextual_sentences(list(doc.sents), j, prev_sentence, relevant_sentences, added_sentences)
            except Exception as e:
                raise Exception(f"Error while checking entity presence and adding contextual sentences: {e}")

            prev_sentence = sentence

        return ' '.join(relevant_sentences)

    def is_entity_present(self, sentence):
        try:
            sentence_text = sentence.text
            sentence_lemmas = {token.lemma_ for token in sentence}
            return self.fixed_entities and (self.entity_pattern.search(sentence_text) or sentence_lemmas.intersection(self.lemmatized_entities))
        except Exception as e:
            raise Exception(f"Error in checking if entity is present: {e}")


    def add_contextual_sentences(self, sentences, current_index, prev_sentence, relevant_sentences, added_sentences):
        if prev_sentence and prev_sentence.text not in added_sentences:
            relevant_sentences.append(prev_sentence.text)
            added_sentences.add(prev_sentence.text)

        current_sentence = sentences[current_index]
        if current_sentence.text not in added_sentences:
            relevant_sentences.append(current_sentence.text)
            added_sentences.add(current_sentence.text)

        if current_index < len(sentences) - 1:
            next_sentence = sentences[current_index + 1]
            if next_sentence.text not in added_sentences:
                relevant_sentences.append(next_sentence.text)
                added_sentences.add(next_sentence.text)

    def measure_reduction(self, original_text: str, reduced_text: str) -> float:
        original_length = len(original_text)
        reduced_length = len(reduced_text)
        reduction_percentage = ((original_length - reduced_length) / original_length) * 100
        return reduction_percentage

    def process_entity_types(self, doc_entities):
        try:
            filtered_entities = []
            def matches_criteria(entity_chunk, entity_labels):
                entity_chunk_lower = entity_chunk.lower()
                if self.fixed_entities:
                    fixed_entities_lower = [entity.lower() for entity in self.fixed_entities]
                    return any(entity in entity_chunk_lower for entity in fixed_entities_lower) and \
                        any(label in self.entity_types for label in entity_labels)
                return any(label in self.entity_types for label in entity_labels)


            for entity_group in doc_entities:
                for entity_data in entity_group:
                    entity1, sentence, entity2, entity_info = entity_data

                    entity1_labels = entity_info.get('entity1_label', '').split(', ')
                    entity2_labels = entity_info.get('entity2_label', '').split(', ')
                    entity1_nn_chunk = entity_info.get('entity1_nn_chunk', '')
                    entity2_nn_chunk = entity_info.get('entity2_nn_chunk', '')
                    entity1_match = matches_criteria(entity1_nn_chunk, entity1_labels)
                    entity2_match = matches_criteria(entity2_nn_chunk, entity2_labels)
                    if entity1_match or entity2_match:
                        filtered_entities.append(entity_data)

            return [filtered_entities]
        except Exception as e:
            raise Exception(f"Error in processing entity types: {e}")


import asyncio
from newsapi import NewsApiClient
from typing import AsyncGenerator
from querent.common.types.collected_bytes import CollectedBytes
from querent.config.collector.collector_config import CollectorConfig, CollectorBackend, NewsCollectorConfig
from querent.collectors.collector_base import Collector
from querent.collectors.collector_factory import CollectorFactory
from querent.common.uri import Uri
from querent.logging.logger import setup_logger

class NewsCollector(Collector):
    """Enhanced class for collecting news articles using NewsAPI with advanced search capabilities and pagination."""

    def __init__(self, config: NewsCollectorConfig):
        self.newsapi = NewsApiClient(api_key=config.api_key)
        self.config = config
        self.logger = setup_logger(__name__, "NewsCollector")

    async def connect(self):
        # Placeholder for future connection logic
        pass

    async def disconnect(self):
        # Placeholder for future disconnection cleanup logic
        pass

    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        current_page = 1
        total_pages = float('inf')  # Assume an infinite number of pages initially

        while current_page <= total_pages:
            try:
                response = self.newsapi.get_everything(
                    q=self.config.query,
                    sources=self.config.sources,
                    domains=self.config.domains,
                    exclude_domains=self.config.exclude_domains,
                    from_param=self.config.from_date,
                    to=self.config.to_date,
                    language=self.config.language,
                    sort_by=self.config.sort_by,
                    page_size=self.config.page_size,
                    page=current_page
                )

                if response['status'] == 'ok':
                    articles = response.get('articles', [])
                    for article in articles:
                        article_data = {
                            "source": article.get('source', {}).get('name'),
                            "author": article.get('author'),
                            "title": article.get('title'),
                            "description": article.get('description'),
                            "url": article.get('url'),
                            "urlToImage": article.get('urlToImage'),
                            "publishedAt": article.get('publishedAt'),
                            "content": article.get('content')
                        }
                        yield CollectedBytes(file=article["title"], data=str(article_data).encode("utf-8"))
                        yield CollectedBytes(file=article["title"], data=None, error=None, eof=True)

                    total_results = response.get('totalResults', 0)
                    total_pages = (total_results + self.config.page_size - 1) // self.config.page_size  # Calculate total pages
                    current_page += 1
                else:
                    self.logger.error(f"News API request failed: {response.get('message')}")
                    break
            except Exception as e:
                self.logger.error(f"Error fetching news articles: {e}")
                yield CollectedBytes(file="Error", data=None, error=e)
                break

class NewsCollectorFactory(CollectorFactory):
    def backend(self) -> CollectorBackend:
        return CollectorBackend.News

    def resolve(self, uri: Uri, config: NewsCollectorConfig) -> Collector:
        return NewsCollector(config)



class IngestedMessages:
    """
    A class to represent the ingested messages.
    """

    def __init__(self, messages):
        """
        Constructor for the IngestedMessages class.

        Parameters
        ----------
        messages : list
            A list of messages.
        """

        self.messages = messages

    def __str__(self):
        """
        Returns a string representation of the IngestedMessages object.

        Returns
        -------
        str
            A string representation of the IngestedMessages object.
        """

        return str(self.messages)

    def __repr__(self):
        """
        Returns a string representation of the IngestedMessages object.

        Returns
        -------
        str
            A string representation of the IngestedMessages object.
        """

        return str(self.messages)



from enum import Enum
from pydantic import BaseModel


class SupportedMetricType(str, Enum):
    GAUGE = "gauge"
    COUNTER = "counter"
    HISTOGRAM = "histogram"


class PrometheusAdapterConfig(BaseModel):
    """
    PrometheusAdapterConfig is the configuration for the Prometheus adapter.

    Attributes:
        port (int): The port on which the Prometheus metrics will be exposed.
        job_name (str): The job name associated with your application in Prometheus.
        labels (list): List of label names for Prometheus metrics.
    """

    job_name: str
    labels: list
    metric_type: SupportedMetricType



from typing import Any, Optional, Dict
from pydantic import BaseModel, validator
from querent.common.types.workflow_config_keys import WorkflowConfigKey
import os


class WorkflowConfig(BaseModel):
    """Workflow configuration."""

    name: str
    id: str
    config: Dict[str, str]
    inner_channel: Optional[Any]
    channel: Optional[Any]
    inner_channel_handler: Optional[Any]
    event_handler: Optional[Any]
    inner_tokens_feader: Optional[Any]
    tokens_feader: Optional[Any]
    

    def __init__(self, config_source=None, **kwargs):
        # if kwargs:
        #     raise ValueError(
        #         "Config values must be provided within a dictionary via 'config_source' parameter."
        #     )

        if config_source:
            config_data = self.load_config(config_source)
            super().__init__(**config_data) 
            for config_key in WorkflowConfigKey:
                key = config_key.value
                if key in config_data:
                    setattr(self, key, config_data[key])

        # else:
        #     raise ValueError("Please pass config")

    @validator("channel", pre=True, allow_reuse=True)
    def validate_channel(cls, value):
        if not hasattr(value, "receive_in_python") or not hasattr(
            value, "send_in_rust"
        ):
            raise ValueError(
                "Invalid type for channel. Must have 'receive_in_python' and 'send_in_rust' functions."
            )
        return value
    
    @validator("tokens_feader", pre=True, allow_reuse=True)
    def validate_tokens_feader(cls, value):
        if not hasattr(value, "receive_tokens_in_python") or not hasattr(
            value, "send_tokens_in_rust"
        ):
            raise ValueError(
                "Invalid type for tokens_feader. Must have 'receive_tokens_in_python' and 'send_tokens_in_rust' functions."
            )
        return value

    @validator("event_handler", pre=True, allow_reuse=True)
    def validate_event_handler(cls, value):
        # value must have handle_event function
        if not hasattr(value, "handle_event"):
            raise ValueError(
                "Invalid type for event_handler. Must have 'handle_event' function."
            )

    @classmethod
    def load_config(cls, config_source) -> dict:
        if isinstance(config_source, dict):
            # If config source is a dictionary, return a dictionary
            cls.config_data = config_source
        else:
            raise ValueError("Invalid config. Must be a valid dictionary")

        env_vars = dict(os.environ)
        cls.config_data.update(env_vars)
        return cls.config_data

    @classmethod
    def get_full_config(cls):
        return cls.config_data

    @classmethod
    def get(cls, key: WorkflowConfigKey, default=None):
        """
        Get a specific configuration value by key.
        Args:
            key (ConfigKey): The key for the configuration value.
            default: The default value to return if the key is not found.
        Returns:
            The configuration value if found, otherwise the default value.
        """
        return cls.config_data.get(key, default)



import asyncio
import psutil

from querent.logging.logger import setup_logger


class ResourceManager:
    def __init__(self, max_allowed_workers=100):
        self.max_allowed_workers = max_allowed_workers
        self.min_allowed_workers = 1
        self.querent_termination_event = asyncio.Event()
        self.logger = setup_logger(__name__, "resource_manager")

    async def get_max_allowed_workers(self):
        return self.max_allowed_workers

    async def get_min_allowed_workers(self):
        return self.min_allowed_workers

    async def get_desired_workers(self):
        # Your logic to calculate the desired number of workers based on system conditions
        # Monitor CPU usage and memory
        cpu_percent = psutil.cpu_percent()
        memory_percent = psutil.virtual_memory().percent

        return int((cpu_percent + memory_percent) / 2)

    async def adjust_max_workers(self, new_max_workers):
        # Adjust the maximum allowed workers dynamically based on system conditions
        if new_max_workers < self.min_allowed_workers:
            new_max_workers = self.min_allowed_workers
        elif new_max_workers > self.max_allowed_workers:
            new_max_workers = self.max_allowed_workers

        self.max_allowed_workers = new_max_workers

    async def is_system_overloaded(self):
        # Check if the system is overloaded based on CPU and memory usage
        cpu_percent = psutil.cpu_percent()
        memory_percent = psutil.virtual_memory().percent

        # Implement your own logic to detect system overload
        if cpu_percent > 90 or memory_percent > 90:
            return True
        else:
            return False

    async def adjust_resources_based_on_load(self):
        # Check system load and adjust resource allocation accordingly
        if await self.is_system_overloaded():
            # Scale down workers if the system is overloaded
            await self.adjust_max_workers(self.max_allowed_workers // 2)
        else:
            # Scale up workers if the system is not overloaded
            await self.adjust_max_workers(self.max_allowed_workers * 2)



from typing import Any


class EventType:
    Graph = "Graph"
    Vector = "Vector"
    Terminate="Terminate"


class EventState:
    def __init__(self, event_type: EventType, timestamp: float, payload: Any, file: str):
        self.event_type = event_type
        self.timestamp = timestamp
        self.payload = payload
        self.file = file
        self.file = file



from unidecode import unidecode
import nltk
from nltk.corpus import words

from querent.processors.async_processor import AsyncProcessor
from querent.logging.logger import setup_logger


class TextProcessor(AsyncProcessor):
    def __init__(self):
        nltk.download("words")
        self.english_vocab = set(w.lower() for w in words.words())
        self.logger = setup_logger(__name__, "TextProcessor")

    async def process_text(self, data: str) -> str:
        try:
            if data is None or data == "":
                return [data]
            text = unidecode(data)

            processed_lines = []
            lines = text.split("\n")
            for index, line in enumerate(lines):
                words_in_line = line.split()
                i = 0
                new_line = ""
                while i < len(words_in_line):
                    word = words_in_line[i]
                    # Check if next word exists
                    if i + 1 < len(words_in_line) and i > 0:
                        next_word = words_in_line[i + 1]
                        prev_word = words_in_line[i - 1]
                        combined_next_word = word + next_word
                        combined_prev_word = prev_word + word
                        # Join words only if the combined form is valid and neither of the individual words is valid
                        if self.is_valid_word(combined_next_word) and not (
                            self.is_valid_word(word) or self.is_valid_word(next_word)
                        ):
                            word = combined_next_word
                            i += 1
                        elif self.is_valid_word(combined_prev_word) and not (
                            self.is_valid_word(word) or self.is_valid_word(prev_word)
                        ):
                            # If individual word formed after joining current word and previous word is valid, then delete the previous word and replace it with new word

                            word = combined_prev_word
                            last_space_index = new_line.rfind(" ")
                            if last_space_index != -1:
                                second_last_space_index = new_line.rfind(
                                    " ", 0, last_space_index
                                )
                                if second_last_space_index == -1:
                                    second_last_space_index = 0
                                new_line = (
                                    new_line[:second_last_space_index]
                                    + new_line[last_space_index:]
                                )
                    new_line += word + " "
                    i += 1

                if (
                    len(line) > 3
                    and index < len(lines) - 1
                    and words_in_line[len(words_in_line) - 1] == "-"
                ):
                    # If the sentence ends with hyphen(-), then check if the word is incomplete, if yes then join them
                    current_word = words_in_line[len(words_in_line) - 2]
                    next_index = lines[index + 1].index(" ")
                    next_word = lines[index + 1][:next_index]
                    new_word = current_word + next_word
                    if self.is_valid_word(new_word):
                        new_line = new_line[: len(new_line) - 3] + next_word
                        lines[index + 1] = lines[index + 1][next_index + 1 :]

                processed_lines.append(new_line.strip())

            return processed_lines
        except Exception as e:
            self.logger.error(f"Exception while processing data {e}")
            return []

    def is_valid_word(self, word):
        return word.lower() in self.english_vocab



import re
import enum
from pathlib import Path
from typing import Optional, Union


class Protocol(enum.Enum):
    Azure = "azure"
    File = "file"
    Gcs = "gs"
    PostgreSQL = "postgresql"
    Ram = "ram"
    S3 = "s3"
    Webscraper = "https"
    Slack = "slack"
    DropBox = "dropbox"
    Github = "github"
    Drive = "drive"
    Email = "email"
    Jira = "jira"
    News = "news"

    def is_azure(self) -> bool:
        return self == Protocol.Azure

    def is_file(self) -> bool:
        return self == Protocol.File

    def is_gcs(self) -> bool:
        return self == Protocol.Gcs

    def is_postgresql(self) -> bool:
        return self == Protocol.PostgreSQL

    def is_ram(self) -> bool:
        return self == Protocol.Ram

    def is_s3(self) -> bool:
        return self == Protocol.S3

    def is_file_storage(self) -> bool:
        return self in {Protocol.File, Protocol.Ram}

    def is_object_storage(self) -> bool:
        return self in {Protocol.Azure, Protocol.S3}

    def is_database(self) -> bool:
        return self == Protocol.PostgreSQL

    def is_webscraper(self) -> bool:
        return self == Protocol.Webscraper

    def is_azure_blob_storage(self) -> bool:
        return self == Protocol.Azure

    def is_slack(self) -> bool:
        return self == Protocol.Slack

    def is_dropbox(self) -> bool:
        return self == Protocol.DropBox

    def is_github(self) -> bool:
        return self == Protocol.Github

    def is_drive(self) -> bool:
        return self == Protocol.Drive

    def is_email(self) -> bool:
        return self == Protocol.Email

    def is_jira(self) -> bool:
        return self == Protocol.Jira
    
    def is_news(self) -> bool:
        return self == Protocol.News


class Uri:
    PROTOCOL_SEPARATOR = "://"
    DATABASE_URI_PATTERN = re.compile(
        r"(?P<before>^.*://.*)(?P<password>:.*@)(?P<after>.*)"
    )

    def __init__(self, uri: str):
        self.uri = uri
        self.protocol_idx = uri.find(self.PROTOCOL_SEPARATOR)
        self.protocol = Protocol(uri[: self.protocol_idx])

    @classmethod
    def from_well_formed(cls, uri: str) -> "Uri":
        protocol_idx = uri.find(cls.PROTOCOL_SEPARATOR)
        protocol_str = uri[:protocol_idx]
        Protocol(protocol_str)
        return cls(uri)

    @property
    def extension(self) -> Optional[str]:
        return Path(self.uri).suffix.lstrip(".")

    @property
    def path(self) -> str:
        return self.uri[self.protocol_idx + len(self.PROTOCOL_SEPARATOR) :]

    def as_str(self) -> str:
        return self.uri

    def is_redacted(self) -> bool:
        return self.protocol.is_database()

    def as_redacted_str(self) -> str:
        if self.protocol.is_database():
            return self.DATABASE_URI_PATTERN.sub(
                r"\g<before>:***redacted***@\g<after>", self.uri
            )
        return self.uri

    def __str__(self) -> str:
        return self.as_redacted_str()

    def __eq__(self, other: Union[str, "Uri"]) -> bool:
        if isinstance(other, Uri):
            return self.uri == other.uri
        return self.uri == other

    def __hash__(self) -> int:
        return hash(self.uri)


# Usage example
if __name__ == "__main__":
    uri_str = "s3://bucket/key"
    uri = Uri(uri_str)
    print(f"Original URI: {uri}")
    print(f"Extension: {uri.extension}")
    print(f"Protocol: {uri.protocol}")
    print(f"Is Azure: {uri.protocol.is_azure()}")
    print(f"Is Redacted: {uri.is_redacted()}")
    print(f"Redacted URI: {uri.as_redacted_str()}")
    print(f"URI == 's3://bucket/key': {uri == 's3://bucket/key'}")



from typing import List, Dict
from pathlib import Path
from enum import Enum


class CollectorErrorKind(Enum):
    NotFound = "not_found"
    Unauthorized = "unauthorized"
    incompatible = "incompatible"
    NotSupported = "not_supported"


class CollectorResolverError(Exception):
    def __init__(self, kind: CollectorErrorKind, message: str):
        super().__init__(message)
        self.kind = kind


class CollectorError(Exception):
    def __init__(self, kind: CollectorErrorKind, source: any):
        self.kind = kind
        self.source = source


class NotFoundError(CollectorError):
    def __init__(self, message: str):
        super().__init__(CollectorErrorKind.NotFound, message)


class UnauthorizedError(CollectorError):
    def __init__(self, message: str):
        super().__init__(CollectorErrorKind.unauthorized, message)


class IncompatibleError(CollectorError):
    def __init__(self, message: str):
        super().__init__(CollectorErrorKind.incompatible, message)



from rdflib.plugins.stores.memory import SimpleMemory


class GraphConfig:
    def __init__(
        self,
        identifier: str,
        format: str = "ttl",
        schema: str = "",
        flush_on_serialize: bool = False,
        store=SimpleMemory(),
        memory_threshold: int = 1024,
    ):
        """
        Initialize the graph configuration.

        Args:
            graph_identifier (str): The identifier for the graph.
            graph_format (str): The format of the graph.
            store (str): The store for the graph.
            memory_threshold (int): this value is in MB.
            schema (str): The schema for the graph.
        """
        self.graph_identifier = identifier
        self.graph_format = format
        self.store = store
        self.memory_threshold = 1024 * 1024 * memory_threshold
        self.flush_on_serialize = flush_on_serialize
        self.schema = schema



from abc import ABC, abstractmethod
from typing import Optional
from enum import Enum
from querent.collectors.collector_base import Collector
from querent.collectors.collector_errors import (
    CollectorResolverError,
    CollectorErrorKind,
)
from querent.config.collector.collector_config import CollectorConfig, CollectorBackend


class CollectorFactory(ABC):
    @abstractmethod
    def backend(self) -> CollectorBackend:
        raise NotImplementedError

    @abstractmethod
    async def resolve(
        self, uri: str, config: CollectorConfig
    ) -> Optional[CollectorBackend]:
        raise NotImplementedError


class UnsupportedCollector(CollectorFactory):
    def __init__(self, backend: CollectorBackend, message: str):
        self.backend = backend
        self.message = message

    async def resolve(self, uri: str, config: CollectorConfig) -> Optional[Collector]:
        raise CollectorResolverError(
            CollectorErrorKind.NotSupported, self.backend, self.message
        )



from typing import Any, Optional
from pydantic import validator
import os
from pydantic import BaseModel

from querent.common.types.engine_config_keys import EngineConfigKey


class EngineConfig(BaseModel):
    """Engine configuration."""

    id: str
    name: str

    num_workers: Optional[int] = 1
    max_retries: Optional[int] = 1
    retry_interval: Optional[int] = 2
    message_throttle_limit: Optional[int] = 1000
    message_throttle_delay: Optional[int] = 1
    # Use Field with allow_mutation=False to specify the type
    inner_channel: Optional[Any]
    channel: Optional[Any]
    logger: str = f"{__name__}.engine_config"
    state_queue: str = f"{__name__}.state_queue"
    workers: str = f"{__name__}.workers"

    def __init__(self, config_source=None, **kwargs):
        # if kwargs:
        #     raise ValueError(
        #         "Config values must be provided within a dictionary via 'config_source' parameter."
        #     )

        if config_source:
            config_data = self.load_config(config_source)
            super().__init__(**config_data)
            for config_key in EngineConfigKey:
                key = config_key.value
                if key in config_data:
                    setattr(self, key, config_data[key])

        # else:
        #     raise ValueError("Please pass config")

    # Custom validator for ChannelCommandInterface
    @validator("channel", pre=True, allow_reuse=True)
    def validate_channel(cls, value):
        if not hasattr(value, "receive_in_python") or not hasattr(
            value, "send_in_rust"
        ):
            raise ValueError(
                "Invalid type for channel. Must have 'receive_in_python' and 'send_in_rust' functions."
            )
        return value

    @classmethod
    def load_config(cls, config_source) -> dict:
        if isinstance(config_source, dict):
            # If config source is a dictionary, return a dictionary
            cls.config_data = config_source
        else:
            raise ValueError("Invalid config. Must be a valid dictionary")

        env_vars = dict(os.environ)
        cls.config_data.update(env_vars)
        return cls.config_data

    @classmethod
    def get_full_config(cls):
        return cls.config_data

    @classmethod
    def get(cls, key: EngineConfigKey, default=None):
        """
        Get a specific configuration value by key.
        Args:
            key (ConfigKey): The key for the configuration value.
            default: The default value to return if the key is not found.
        Returns:
            The configuration value if found, otherwise the default value.
        """
        return cls.config_data.get(key, default)



"""Querent

Querent is an async data dynamo and inference engine for Python. It is designed to
be used as a library in Python applications that need to perform inference on
data in a scalable and efficient manner. Querent is built on top of the Python
asyncio library and is designed to be used in an async manner.

Querent is the Python Asyncio Inference Engine by Querent AI.
Copyright (c) 2023 by Querent AI.

"""
import asyncio
import signal
from typing import List, Awaitable
from querent.core.base_engine import BaseEngine
from querent.querent.resource_manager import ResourceManager
from querent.querent.auto_scaler import AutoScaler
from querent.logging.logger import setup_logger


class Querent:
    def __init__(
        self,
        querenters: List[BaseEngine],
        resource_manager: ResourceManager,
    ):
        self.logger = setup_logger(__name__, "querent")
        self.resource_manager = resource_manager
        self.querenters = querenters
        self.auto_scaler = AutoScaler(self.resource_manager, querenters)

        # Create an event to handle termination requests
        self.querent_termination_event = resource_manager.querent_termination_event

    async def start(self):
        try:
            self.logger.info("Starting Querent")

            # Start the auto-scaler
            auto_scale_task = asyncio.create_task(self.auto_scaler.start())

            # Start handling signals
            #self.setup_signal_handlers()

            # Start the tasks above and wait for them to finish
            await asyncio.gather(auto_scale_task, self.wait_for_termination())

        except Exception as e:
            self.logger.error(f"An error occurred during Querent execution: {e}")
            await self.graceful_shutdown()
        finally:
            await self.graceful_shutdown()
            self.logger.info("Querent stopped")

    async def graceful_shutdown(self):
        self.logger.info("Initiating graceful shutdown of Querent")

        # Stop the auto-scaler and querenters gracefully
        await self.auto_scaler.stop()

        self.logger.info("Querent stopped gracefully")

    def setup_signal_handlers(self):
        for sig in [signal.SIGINT, signal.SIGTERM]:
            loop = asyncio.get_event_loop()
            loop.add_signal_handler(sig, self.handle_signal)

    def handle_signal(self):
        try:
            shutdown_task = asyncio.create_task(self.graceful_shutdown())
            asyncio.run(shutdown_task)
        except Exception as e:
            print(f"Error during graceful shutdown: {str(e)}")

    async def wait_for_termination(self) -> Awaitable[None]:
        # Wait for the termination event to be set, indicating graceful shutdown
        await self.querent_termination_event.wait()



from typing import Optional
from querent.collectors.azure.azure_collector import AzureCollectorFactory
from querent.collectors.email.email_collector import EmailCollectorFactory
from querent.collectors.gcs.gcs_collector import GCSCollectorFactory
from querent.collectors.aws.aws_collector import AWSCollectorFactory
from querent.collectors.fs.fs_collector import FSCollectorFactory
from querent.collectors.jira.jira_collector import JiraCollectorFactory
from querent.collectors.webscaper.web_scraper_collector import WebScraperFactory
from querent.collectors.slack.slack_collector import SlackCollectorFactory
from querent.collectors.dropbox.dropbox_collector import DropBoxCollectorFactory
from querent.collectors.github.github_collector import GithubCollectorFactory
from querent.collectors.drive.google_drive_collector import DriveCollectorFactory
from querent.collectors.news.news_collector import NewsCollectorFactory
from querent.config.collector.collector_config import CollectorConfig, CollectorBackend
from querent.collectors.collector_base import Collector
from querent.collectors.collector_errors import (
    CollectorResolverError,
    CollectorErrorKind,
)
from querent.common.uri import Protocol, Uri


class CollectorResolver:
    def __init__(self):
        self.collector_factories = {
            CollectorBackend.LocalFile: FSCollectorFactory(),
            CollectorBackend.S3: AWSCollectorFactory(),
            CollectorBackend.WebScraper: WebScraperFactory(),
            CollectorBackend.Gcs: GCSCollectorFactory(),
            CollectorBackend.AzureBlobStorage: AzureCollectorFactory(),
            CollectorBackend.Slack: SlackCollectorFactory(),
            CollectorBackend.DropBox: DropBoxCollectorFactory(),
            CollectorBackend.Github: GithubCollectorFactory(),
            CollectorBackend.Drive: DriveCollectorFactory(),
            CollectorBackend.Email: EmailCollectorFactory(),
            CollectorBackend.Jira: JiraCollectorFactory(),
            CollectorBackend.News: NewsCollectorFactory(),
            # Add other collector factories as needed
        }

    def resolve(self, uri: Uri, config: CollectorConfig) -> Optional[Collector]:
        backend = config.backend
        backendFromUri = self._determine_backend(uri.protocol)

        if backend != backendFromUri:
            raise CollectorResolverError(
                CollectorErrorKind.NotSupported,
                "Backend in the config does not match the backend the protocol",
            )

        if backend in self.collector_factories:
            factory = self.collector_factories[backend]
            return factory.resolve(uri, config)
        else:
            raise CollectorResolverError(
                CollectorErrorKind.NotSupported, "Unsupported backend"
            )

    def _determine_backend(self, protocol: Protocol) -> CollectorBackend:
        if protocol.is_file_storage():
            return CollectorBackend.LocalFile
        elif protocol.is_s3():
            return CollectorBackend.S3
        elif protocol.is_webscraper():
            return CollectorBackend.WebScraper
        elif protocol.is_gcs():
            return CollectorBackend.Gcs
        elif protocol.is_webscraper():
            return CollectorBackend.WebScraper
        elif protocol.is_azure_blob_storage():
            return CollectorBackend.AzureBlobStorage
        elif protocol.is_slack():
            return CollectorBackend.Slack
        elif protocol.is_dropbox():
            return CollectorBackend.DropBox
        elif protocol.is_github():
            return CollectorBackend.Github
        elif protocol.is_drive():
            return CollectorBackend.Drive
        elif protocol.is_email():
            return CollectorBackend.Email
        elif protocol.is_jira():
            return CollectorBackend.Jira
        elif protocol.is_news():
            return CollectorBackend.News
        else:
            raise CollectorResolverError(
                CollectorErrorKind.NotSupported, "Unknown backend"
            )



from pyshacl import validate
from typing import Dict, Union
from rdflib import RDF, RDFS, Namespace, XSD
import rdflib
from querent.config.graph_config import GraphConfig
from querent.graph.graph import QuerentGraph
from querent.graph.subject import Subject

from querent.graph.utils import URI, BNode, Literal, LiteralType

SH = Namespace("http://www.w3.org/ns/shacl#")


class SHACL(QuerentGraph):
    def __init__(self, config: GraphConfig):
        super().__init__(config)
        self._class_nodes: Dict[str, Union[URI, BNode]] = {}
        self.bind("sh", SH)

    def add_ontology(self, onto_graph):
        converter = SHACLOntoConverter(self._class_nodes)
        self.graph += converter.convert_ontology(onto_graph)._g
        self._class_nodes.update(converter.class_nodes)

    def validate(self, data_graph, onto_graph=None, inference=None):
        if isinstance(data_graph, QuerentGraph):
            data_graph = data_graph._g
        if isinstance(onto_graph, QuerentGraph):
            onto_graph = onto_graph._g
        if onto_graph and inference:
            conforms, results_graph, results_text = validate(
                data_graph + onto_graph, shacl_graph=self.graph, inference=inference
            )
        else:
            conforms, results_graph, results_text = validate(
                data_graph, shacl_graph=self.graph
            )
        results_graph = QuerentGraph(results_graph)
        return conforms, results_graph


class SHACLOntoConverter:
    def __init__(self, class_nodes: Dict[str, Union[URI, BNode]] = None):
        super().__init__()
        self.onto_graph: rdflib.QuerentGraph = None
        self.graph = QuerentGraph()
        self._class_nodes = class_nodes if class_nodes else {}
        self.graph.bind("sh", SH)

    def cnode(self, class_):
        class_ = str(class_)
        if class_ not in self._class_nodes:
            self._class_nodes[class_] = BNode()
        return self._class_nodes[class_]

    @property
    def class_nodes(self):
        return self._class_nodes

    def convert_ontology(self, onto_graph: QuerentGraph) -> QuerentGraph:
        self.onto_graph = (
            onto_graph._g if isinstance(onto_graph, QuerentGraph) else onto_graph
        )
        # build shacl property-based NodeShape for non-domain-referenced properties
        self._convert_ontology_non_referenced_property()
        # build shacl class-based NodeShape
        # build shacl property path-based blank node for domain-referenced properties
        # for path-based property,
        # we can only add sh:datatype/sh:class based on rdfs:range
        self._convert_ontology_class_node_shape()

        # build shacl property based on owl restriction
        # we can add cardinality to it
        self._convert_ontology_owl_restriction()
        return self.graph

    def _property_shape(self, property_: Union[rdflib.URIRef, str, URI]):
        if isinstance(property_, URI):
            property_ = property_.value
        property_ = rdflib.URIRef(property_)
        p_shape = Subject(BNode())
        p_shape.add_property(URI("sh:path"), URI(property_))
        ranges = list(self.onto_graph.objects(property_, RDFS.range))
        self._build_property_ranges(p_shape, ranges)
        return p_shape

    def _build_property_ranges(self, p_shape, ranges):
        if not ranges:
            return
        if len(ranges) == 1:
            range_ = ranges[0]
            type_ = URI("sh:datatype") if self._is_datatype(range_) else URI("sh:class")
            p_shape.add_property(type_, URI(ranges[0]))
        else:
            or_list = []
            for range_ in ranges:
                range_subject = Subject(BNode())
                type_ = (
                    URI("sh:datatype") if self._is_datatype(range_) else URI("sh:class")
                )
                range_subject.add_property(type_, URI(range_))
                or_list.append(range_subject)
            p_shape.add_property(URI("sh:or"), self._add_list(or_list))

    @staticmethod
    def _is_datatype(uri: rdflib.URIRef):
        if isinstance(uri, rdflib.BNode):
            return False
        return uri.startswith(str(XSD)) or uri.startswith(str(RDF))

    def _convert_ontology_non_referenced_property(self):
        for (p,) in self.onto_graph.query(
            """
          SELECT ?p
          WHERE {
            { ?p a rdfs:Property }
            UNION
            { ?p a owl:ObjectProperty }
            UNION
            { ?p a owl:DatatypeProperty }
            FILTER NOT EXISTS { ?p rdfs:domain ?c }
            FILTER NOT EXISTS { ?c owl:onProperty ?p }
          }
        """
        ):
            p_shape = self._property_shape(p)
            p_shape.add_property(URI("rdf:type"), URI("sh:PropertyShape"))
            self.graph.add_subject(p_shape)

    def _convert_ontology_class_node_shape(self):
        for (c,) in self.onto_graph.query(
            """
          SELECT DISTINCT ?s WHERE {
            {?s a owl:Class}
            UNION
            {?s a rdfs:Class}
            UNION
            {?p rdfs:domain ?s}
            UNION
            {?s rdfs:subClassOf|owl:equivalentClass ?o}
            UNION
            {?os rdfs:subClassOf|owl:equivalentClass ?s FILTER NOT EXISTS {?s a owl:Restriction}}
            FILTER isURI(?s)
          }
        """
        ):
            """
            [ a sh:NodeShape ;
              sh:targetClass ?s ;
              sh:property [ sh:path p ; sh:class class ;] ;]
            """
            node_subject = self._class_shape(c)
            for p in self.onto_graph.subjects(RDFS.domain, c):
                property_subject = self._property_shape(p)
                node_subject.add_property(URI("sh:property"), property_subject)
            self.graph.add_subject(node_subject)

    def _class_shape(self, c):
        node_subject = Subject(self.cnode(c))
        node_subject.add_property(URI("rdf:type"), URI("sh:NodeShape"))
        node_subject.add_property(URI("sh:targetClass"), URI(c))
        return node_subject

    def _convert_ontology_owl_restriction(self):
        for c, p, exact, min_, max_ in self.onto_graph.query(
            """
          SELECT ?c ?p ?exact ?min ?max
          WHERE {
            ?c rdfs:subClassOf|owl:equivalentClass ?res .
            ?res a owl:Restriction ;
                 owl:onProperty ?p .
            OPTIONAL {?res owl:cardinality ?exact}
            OPTIONAL {?res owl:minCardinality ?min}
            OPTIONAL {?res owl:maxCardinality ?max}
          }
        """
        ):
            node_subject = self._class_shape(c)
            property_shape = self._property_shape(p)
            if exact:
                property_shape.add_property(
                    URI("sh:count"), Literal(str(exact), type_=LiteralType.integer)
                )
            if min_:
                property_shape.add_property(
                    URI("sh:minCount"), Literal(str(min_), type_=LiteralType.integer)
                )
            if max_:
                property_shape.add_property(
                    URI("sh:maxCount"), Literal(str(max_), type_=LiteralType.integer)
                )
            ranges = []
            for (r,) in self.onto_graph.query(
                """
              SELECT ?r
              WHERE {
                ?c rdfs:subClassOf|owl:equivalentClass ?res .
                ?res a owl:Restriction ;
                     owl:onProperty ?p ;
                     owl:allValuesFrom|owl:someValuesFrom|owl:hasValue ?r
              }
            """,
                initBindings={"c": c, "p": p},
            ):
                ranges.append(r)
            self._build_property_ranges(property_shape, ranges)
            node_subject.add_property(URI("sh:property"), property_shape)
            self.graph.add_subject(node_subject)

    def _add_list(self, list_):
        """
        (:a :b :c) ===
        [ rdf:first :a ; rdf:rest [ rdf:first :b ; rdf:rest [ rdf:first :c ; rdf:rest rdf:nil]]]
        """
        if not list_:
            return URI("rdf:nil")
        head = Subject(BNode())
        head.add_property(URI("rdf:first"), list_[0])
        head.add_property(URI("rdf:rest"), self._add_list(list_[1:]))
        return head



from enum import Enum
from typing import Any, List, Optional, Union, Dict
from pydantic import BaseModel, Field, validator
import os
from querent.common.types.collector_config_keys import CollectorConfigKey

from querent.channel.channel_interface import ChannelCommandInterface


class CollectorBackend(str, Enum):
    LocalFile = "localfile"
    WebScraper = "webscraper"
    S3 = "s3"
    Gcs = "gs"
    AzureBlobStorage = "azure"
    Slack = "slack"
    DropBox = "dropbox"
    Github = "github"
    Drive = "drive"
    Email = "email"
    Jira = "jira"
    News = "news"

class CollectorConfig(BaseModel):
    backend: CollectorBackend
    # Use Field with allow_mutation=False to specify the type
    channel: Optional[Any]
    id: str
    name: str
    uri: Optional[Any]
    config: Dict[str, str]
    inner_channel: Optional[Any]
    config_source: Optional[Any]

    def __init__(self, config_source=None, **kwargs):

        if config_source:
            config_data = self.load_config(config_source)

            if "uri" not in config_data or not config_data["uri"]:
                config_data["uri"] = uri_backend_mapping[config_data.get("backend")]

            super().__init__(**config_data)
            self.config_source = config_source
            for config_key in CollectorConfigKey:
                key = config_key.value
                if key in config_data:
                    setattr(self, key, config_data[key])

        # else:
        #     raise ValueError("Please pass config")
    
    def resolve(self):
        if self.config_source and isinstance(self.config_source, dict):
            backend_type = self.config_source.get('backend')
            if backend_type in colectorconfig_factories:
                collector_class = colectorconfig_factories[backend_type]
                return collector_class(self.config_source)
            else:
                raise ValueError(f"Unsupported backend type: {backend_type}")
        else:
            raise ValueError("Invalid or missing config source")

    # Custom validator for ChannelCommandInterface
    @validator("channel", pre=True, allow_reuse=True)
    def validate_channel(cls, value):
        if not hasattr(value, "receive_in_python") or not hasattr(
            value, "send_in_rust"
        ):
            raise ValueError(
                "Invalid type for channel. Must have 'receive_in_python' and 'send_in_rust' functions."
            )
        return value

    @classmethod
    def load_config(cls, config_source) -> dict:
        if isinstance(config_source, dict):
            # If config source is a dictionary, return a dictionary
            cls.config_data = config_source
        else:
            raise ValueError("Invalid config. Must be a valid dictionary")

        env_vars = dict(os.environ)
        cls.config_data.update(env_vars)
        return cls.config_data

    @classmethod
    def get_full_config(cls):
        return cls.config_data

    @classmethod
    def get(cls, key: CollectorConfigKey, default=None):
        """
        Get a specific configuration value by key.
        Args:
            key (ConfigKey): The key for the configuration value.
            default: The default value to return if the key is not found.
        Returns:
            The configuration value if found, otherwise the default value.
        """
        return cls.config_data.get(key, default)


class FSCollectorConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.LocalFile
    id: str
    root_path: str
    chunk_size: str = "1024"
    channel: Any

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)

    # Custom validator for ChannelCommandInterface
    @validator("channel", pre=True, allow_reuse=True)
    def validate_channel(cls, value):
        if (
            value is None
            or not hasattr(value, "receive_in_python")
            or not hasattr(value, "send_in_rust")
        ):
            raise ValueError(
                "Invalid type for channel. Must have 'receive_in_python' and 'send_in_rust' functions."
            )
        return value


class AzureCollectConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.AzureBlobStorage
    id: str
    connection_string: str
    account_url: str
    credentials: str
    container: str
    prefix: str
    chunk_size: str = "1024"

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config)
 
        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)


class S3CollectConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.S3
    id: str
    bucket: str
    region: str
    access_key: str
    secret_key: str
    chunk: str = "1024"

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)


class GcsCollectConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.Gcs
    id: str
    bucket: str
    credentials: str
    chunk: str = "1024"

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)


class SlackCollectorConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.Slack
    id: str
    cursor: Optional[str]
    include_all_metadata: Optional[bool]
    inclusive: Optional[bool]
    latest: Optional[str]
    limit: Optional[int]
    channel_name: str
    access_token: str

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)


class DropboxConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.DropBox
    id: str
    dropbox_app_key: str
    dropbox_app_secret: str
    folder_path: str
    chunk_size: str = "1024"
    dropbox_refresh_token: str

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)


class GithubConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.Github
    id: str
    github_username: str
    github_access_token: str
    repository: str

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)


class WebScraperConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.WebScraper
    id: str
    website_url: str = Field(..., description="The URL of the website to scrape.")

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)


class DriveCollectorConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.Drive
    id: str
    drive_refresh_token: str
    drive_token: str
    drive_scopes: str
    drive_client_id: str
    drive_client_secret: str
    chunk_size: str = "1024"
    specific_file_type: Optional[str] = None
    folder_to_crawl: Optional[str] = None

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)


class EmailCollectorConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.Email
    id: str
    imap_server: str
    imap_port: int
    imap_username: str
    imap_password: str
    imap_folder: str
    imap_keyfile: Optional[str] = None
    imap_certfile: Optional[str] = None

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)


class JiraCollectorConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.Jira
    id: str
    jira_server: str
    jira_username: str
    jira_project: str
    jira_query: str
    jira_password: Optional[str] = None
    jira_api_token: Optional[str] = None
    jira_start_at: Optional[int] = 0
    jira_max_results: Optional[int] = 50
    jira_fields: Optional[Union[str, List[str]]] = ("*all",)
    jira_expand: Optional[str] = None
    jira_keyfile: Optional[str] = None
    jira_certfile: Optional[str] = None
    jira_verify: Optional[bool] = True

    def __init__(self, config_source=None, **kwargs):
        if config_source and "config" in config_source:
            extended_config = config_source["config"]
            config_source.update(extended_config) 

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)

from querent.config.collector.collector_config import CollectorConfig, CollectorBackend

class NewsCollectorConfig(CollectorConfig):
    backend: CollectorBackend = CollectorBackend.News
    id: str
    api_key: str
    query: str
    sources: str = None
    domains: str = None
    exclude_domains: str = None
    from_date: str
    to_date: str
    language: str = "en"
    sort_by: str = "publishedAt"
    page_size: int = 100
    page: int = 1

    def __init__(self, config_source=None, **kwargs):
        if config_source is None:
            config_source = {}
        elif "config" in config_source:
            extended_config = config_source.get("config")
            config_source.update(extended_config)

        super().__init__(config_source=config_source, **kwargs)
        for key, value in extended_config.items():
            if hasattr(self, key):
                setattr(self, key, value)
        
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)


colectorconfig_factories = {
        CollectorBackend.LocalFile: FSCollectorConfig,
        CollectorBackend.Drive: DriveCollectorConfig,
        CollectorBackend.AzureBlobStorage: AzureCollectConfig,
        CollectorBackend.DropBox: DropboxConfig,
        CollectorBackend.Email: EmailCollectorConfig,
        CollectorBackend.Gcs: GcsCollectConfig,
        CollectorBackend.Github: GithubConfig,
        CollectorBackend.Jira: JiraCollectorConfig,
        CollectorBackend.S3: S3CollectConfig,
        CollectorBackend.WebScraper: WebScraperConfig,
        CollectorBackend.Slack: SlackCollectorConfig,
        CollectorBackend.News: NewsCollectorConfig,
    }


uri_backend_mapping = {
    "localfile": "file://",
    "webscraper": "https://",
    "s3": "s3://",
    "gs": "gs://",
    "azure": "azure://",
    "slack": "slack://",
    "dropbox": "dropbox://",
    "github": "github://",
    "drive": "drive://",
    "email": "email://",
    "jira": "jira://",
    "news": "news://",
}


import psycopg2
from psycopg2 import sql


class DatabaseConnection:
    def __init__(self, dbname, user, password, host, port):
        self.conn = None
        try:
            self.conn = psycopg2.connect(dbname=dbname, user=user, password=password, host=host, port=port)
            self.conn.autocommit = True
        except Exception as e:
            raise("Unable to connect to the database.")

    def insert_graph_event(self, data):
        if self.conn is not None:
            try:
                with self.conn.cursor() as cur:
                    query = sql.SQL("""INSERT INTO graph_event (subject, subject_type, object, object_type, predicate, predicate_type, sentence, document_id)
                                       VALUES (%s, %s, %s, %s, %s, %s, %s, %s);""")
                    cur.execute(query, (data['subject'], data['subject_type'], data['object'], data['object_type'], data['predicate'], data['predicate_type'], data['sentence'], data['document_id']))
            except Exception as e:
                raise("An error occurred while inserting the data:")

    def close(self):
        if self.conn is not None:
            self.conn.close()


"""
    Querent AI: The Asynchronous Data Dynamo and Graph Neural Network Catalyst
"""

from setuptools import setup, find_packages

requirements = [
    "cachetools==5.3.1",
    "aiohttp==3.9.2",
    "attrs==23.1.0",
    "beautifulsoup4==4.12.2",
    "boto3==1.26.146",
    "botocore==1.29.146",
    "bs4==0.0.1",
    "faiss-cpu==1.7.4",
    "gensim==4.3.2",
    "hdbscan==0.8.33",
    "jira==3.5.0",
    "jmespath==1.0.1",
    "joblib==1.2.0",
    "json5==0.9.14",
    "jsonmerge==1.9.0",
    "jsonschema==4.17.3",
    "kombu==5.2.4",
    "llama_cpp_python==0.2.15",
    "lxml==4.9.2",
    "newspaper3k==0.2.8",
    "nltk==3.8.1",
    "numpy==1.24.3",
    "Pillow==10.0.1",
    "pydantic==1.10.8",
    "PyJWT==2.4.0",
    "pytest==7.3.2",
    "python-dotenv==1.0.0",
    "redis==4.5.5",
    "regex==2023.5.5",
    "sentence-transformers==2.2.2",
    "spacy==3.7.2",
    "uvicorn==0.22.0",
    "slack-sdk==3.26.1",
    "pylint==2.17.4",
    "pytest-cov==4.1.0",
    "pytest-mock==3.11.1",
    "tensorflow==2.14.0",
    "transformers==4.36.0",
    "pypdf==4.0.2",
    "asyncio==3.4.3",
    "aiofiles==23.2.1",
    "prometheus-client==0.17.1",
    "rdflib==7.0.0",
    "pytest-asyncio==0.23.2",
    "pyshacl==0.25.0",
    "google-cloud-storage==2.14.0",
    "pymupdf==1.23.7",
    "pydub==0.25.1",
    "SpeechRecognition==3.10.1",
    "pytesseract==0.3.10",
    "pytextract==2.0.1",
    "pandas==2.1.4",
    "python-pptx==0.6.23",
    "tika==2.6.0",
    "openpyxl==3.1.2",
    "coverage==7.3.3",
    "azure-storage-blob==12.19.0",
    "moviepy==1.0.3",
    "python-docx==1.1.0",
    "dropbox==11.36.2",
    "requests==2.31.0",
    "google-api-python-client==2.105.0",
    "requests_html==0.10.0",
    "rapidocr-onnxruntime==1.3.9",
    "pybase64==1.3.1",
    "pdfminer==20191125",
    "unidecode==1.3.7",
    "psutil==5.9.8",
    "langchain-community==0.0.16",
    "langchain==0.1.4",
    "openai==1.12.0",
    "newsapi-python==0.2.7",
]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="querent",
    version="1.1.0",
    author="Querent AI",
    description="The Asynchronous Data Dynamo and Graph Neural Network Catalyst",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Querent-ai/querent-ai",
    project_urls={
        "Documentation": "https://github.com/Querent-ai/querent-ai/docs",
        "Issue Tracker": "https://github.com/Querent-ai/querent-ai/issues",
    },
    keywords=[
        "Graph Neural Network",
        "Scalability",
        "Data-Driven Insights",
        "GNN",
        "Async",
        "Knowledge Graphs",
        "KG",
        "Large Language Models",
        "asyncio",
        "Insights",
        "aysnchronous",
        "LLM",
        "transformers",
        "pytorch",
        "Llama-index",
        "AI",
        "Artificial Intelligence",
        "Neo4j",
        "Queues",
        "QuiAssisstant",
        "Collectors",
        "Data",
        "Data Science",
        "Data Engineering",
        "Data Analysis",
        "Data Analytics",
        "News",
        "NLP",
        "Natural Language Processing",
        "Text",
        "Text Analysis",
        "Deep Learning",
        "Graphs",
        "Graph Theory",
        "Graph Algorithms",
        "Graph Analytics",
        "Graph Databases",
        "Graph Processing",
        "Graph Mining",
        "Graph Neural Networks",
        "GNN",
        "GNNs",
        "Graph Neural Network",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Telecommunications Industry",
        "Intended Audience :: Other Audience",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: User Interfaces",
        "Topic :: Software Development :: Widget Sets",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: XML",
        "Topic :: Text Processing :: Markup :: Markdown",
        "Topic :: Utilities",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Typing :: Typed",
    ],
    python_requires=">=3.9, <4",
    packages=find_packages(exclude=("tests", "tests.*")),
    install_requires=requirements,
    license="Business Source License 1.1",
)



from typing import AsyncGenerator, List
from io import BytesIO
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.ingested_images import IngestedImages
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.ingestors.base_ingestor import BaseIngestor
from querent.ingestors.ingestor_factory import IngestorFactory
from querent.logging.logger import setup_logger
from querent.processors.async_processor import AsyncProcessor
from querent.common import common_errors
import uuid
import pypdf
from PIL import Image
import io

import pybase64
import pytesseract


class PdfIngestorFactory(IngestorFactory):
    SUPPORTED_EXTENSIONS = {"pdf"}

    async def supports(self, file_extension: str) -> bool:
        return file_extension.lower() in self.SUPPORTED_EXTENSIONS

    async def create(
        self, file_extension: str, processors: List[AsyncProcessor]
    ) -> BaseIngestor:
        if not await self.supports(file_extension):
            return None
        return PdfIngestor(processors)


class PdfIngestor(BaseIngestor):
    def __init__(self, processors: List[AsyncProcessor]):
        super().__init__(IngestorBackend.PDF)
        self.processors = processors
        self.logger = setup_logger(__name__, "PdfIngestor")

    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedTokens or str, None]: # type: ignore
        current_file = None
        collected_bytes = b""

        try:
            async for chunk_bytes in poll_function:
                if chunk_bytes.is_error() or chunk_bytes.is_eof():
                    # report to metrics
                    continue

                if current_file is None:
                    current_file = chunk_bytes.file
                elif current_file != chunk_bytes.file:
                    # we have a new file, process the old one
                    async for page_text in self.extract_and_process_pdf(
                        CollectedBytes(file=current_file, data=collected_bytes)
                    ):
                        yield page_text
                    collected_bytes = b""
                    current_file = chunk_bytes.file
                    yield IngestedTokens(
                        file=current_file,
                        data=None,
                        error=None,
                    )
                collected_bytes += chunk_bytes.data
        except Exception as e:
            # at the queue level, we can sample out the error
            yield IngestedTokens(file=current_file, data=None, error=f"Exception: {e}")
        finally:
            # process the last file
            try:
                async for page_text in self.extract_and_process_pdf(
                    CollectedBytes(file=current_file, data=collected_bytes)
                ):
                    yield page_text
                yield IngestedTokens(file=current_file, data=None, error=None)
            except Exception as exc:
                yield IngestedTokens(
                    file=current_file,
                    data=None,
                    error=f"Exception: {exc}",
                )

    async def extract_and_process_pdf(
        self, collected_bytes: CollectedBytes
    ) -> AsyncGenerator[IngestedTokens, None]:
        try:
            path = BytesIO(collected_bytes.data)
            loader = pypdf.PdfReader(path)

            for page_num, page in enumerate(loader.pages):
                text = page.extract_text()
                if not text:
                    continue

                processed_text = await self.process_data(text)

                # Yield processed text as IngestedTokens
                yield IngestedTokens(
                    file=collected_bytes.file,
                    data=processed_text,
                    error=collected_bytes.error,
                )
                # async for image_result in self.extract_images_and_ocr(
                #     page,
                #     page_num,
                #     processed_text,
                #     collected_bytes.data,
                #     collected_bytes.file,
                # ):
                #     yield image_result

        except TypeError as exc:
            self.logger.error(f"Exception while extracting pdf {exc}")
            raise common_errors.TypeError(
                f"Getting type error on this file {collected_bytes.file}"
            ) from exc

        except Exception as exc:
            self.logger.error(f"Exception while extracting pdf {exc}")
            raise common_errors.UnknownError(
                f"Getting unknown error while handling this file: {collected_bytes.file} error - {exc}"
            ) from exc

    async def extract_images_and_ocr(self, page, page_num, text, data, file_path):
        try:
            for image_path in page.images:
                ocr = await self.get_ocr_from_image(image_path)
                yield IngestedImages(
                    file=file_path,
                    image=pybase64.b64encode(data),
                    image_name=uuid.uuid4(),
                    page_num=page_num,
                    text=text,
                    coordinates=None,
                    ocr_text=ocr,
                )
        except Exception as e:
            self.logger.error(f"Error extracting images and OCR: {e}")
            yield IngestedImages(
                file=file_path,
                image=pybase64.b64encode(data),
                image_name=uuid.uuid4(),
                page_num=page_num,
                text=text,
                coordinates=None,
                ocr_text=None,
                error=f"Exception:{e}",
            )

    async def get_ocr_from_image(self, image):
        """Implement this to return ocr text of the image"""
        try:
            image = Image.open(io.BytesIO(image.data))
            text = pytesseract.image_to_string(image)
        except Exception as e:
            print("Exception-{e}")
            raise e
        return str(text).encode("utf-8").decode("unicode_escape")

    async def process_data(self, text: str) -> str:
        if self.processors is None or len(self.processors) == 0:
            return [text]
        try:
            processed_data = text
            for processor in self.processors:
                processed_data = await processor.process_text(processed_data)
            return processed_data
        except Exception as e:
            self.logger.error(f"Error while processing text: {e}")



from collections import defaultdict
from typing import Any, List

import requests
from querent.callback.event_callback_interface import EventCallbackInterface
from querent.common.types.querent_event import EventState, EventType


class EventCallbackDispatcher:
    def __init__(self, retries=3):
        self.callbacks: dict[EventType, List[EventCallbackInterface]] = defaultdict(
            list
        )

        self.webhooks: dict[EventType, List[str]] = defaultdict(list)
        self.retries = retries

    def register_callback(
        self, event_type: EventType, callback: EventCallbackInterface
    ):
        """
        Register a callback.
        Args:
            callback (EventCallbackInterface): An event callback instance.
        """
        self.callbacks[event_type].append(callback)

    async def dispatch_event(self, event_type: EventType, event_data: EventState):
        """
        Dispatch an event to all registered callbacks.
        Args:
            event_data (Any): Data associated with the event.
        """
        for callback in self.callbacks[event_type]:
            callback.handle_event(event_type, event_data)

    def register_webhook(self, event_type: EventType, webhook: str):
        """
        Register a webhook.
        Args:
            webhook (str): A webhook.
        """
        self.webhooks[event_type].append(webhook)

    async def dispatch_webhook(self, event_type: EventType, event_data: EventState):
        """
        Dispatch an event to all registered webhooks.
        Args:
            event_data (Any): Data associated with the event.
        """
        for webhook_url in self.webhooks[event_type]:
            # Assuming event_data is a dictionary that you want to send as JSON
            try:
                attempt = 0
                while attempt <= self.retries:
                    response = requests.post(webhook_url, json=event_data)
                    if response.status_code == 200:
                        print(f"Webhook to {webhook_url} was successfully dispatched.")
                        break
                    else:
                        print(
                            f"Failed to dispatch webhook to {webhook_url}. Status code: {response.status_code}"
                        )
                        attempt += 1
            except Exception as e:
                print(f"Error when sending webhook to {webhook_url}: {str(e)}")



import asyncio
import json
from querent.core.transformers.fixed_entities_set_opensourcellm import Fixed_Entities_LLM
from querent.kg.ner_helperfunctions.fixed_predicate import FixedPredicateExtractor
from querent.config.core.gpt_llm_config import GPTConfig
from querent.core.transformers.bert_ner_opensourcellm import BERTLLM
from querent.common.types.ingested_images import IngestedImages
from querent.kg.rel_helperfunctions.openai_functions import FunctionRegistry
from querent.common.types.querent_event import EventState, EventType
from querent.core.base_engine import BaseEngine
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.ingested_messages import IngestedMessages
from querent.common.types.ingested_code import IngestedCode
from querent.common.types.querent_queue import QuerentQueue
from querent.kg.rel_helperfunctions.embedding_store import EmbeddingStore
from typing import Any, List, Tuple
from querent.kg.rel_helperfunctions.triple_to_json import TripleToJsonConverter
from querent.logging.logger import setup_logger
from querent.config.core.llm_config import LLM_Config
from openai import OpenAI
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    wait_fixed
)
from dotenv import load_dotenv, find_dotenv
import json

_ = load_dotenv(find_dotenv())


class GPTLLM(BaseEngine):
    def __init__(
        self,
        input_queue: QuerentQueue,
        config: GPTConfig
    ):
        self.logger = setup_logger(__name__, "OPENAILLM")
        try:
            super().__init__(input_queue)
            llm_config = LLM_Config(
            ner_model_name=config.ner_model_name,
            enable_filtering=config.enable_filtering,
            filter_params={
                'score_threshold': config.filter_params['score_threshold'],
                'attention_score_threshold': config.filter_params['attention_score_threshold'],
                'similarity_threshold': config.filter_params['similarity_threshold'],
                'min_cluster_size': config.filter_params['min_cluster_size'],
                'min_samples': config.filter_params['min_samples'],
                'cluster_persistence_threshold':config.filter_params['cluster_persistence_threshold']
            },
            sample_entities = config.sample_entities,
            fixed_entities = config.fixed_entities,
            skip_inferences= True,
            is_confined_search = config.is_confined_search,
            huggingface_token = config.huggingface_token)
            self.is_confined_search = config.is_confined_search
            self.fixed_relationships = config.fixed_relationships
            self.sample_relationships = config.sample_relationships
            if self.fixed_relationships and not self.sample_relationships:
                raise ValueError("If specific predicates are provided, their types should also be provided.")
            if self.fixed_relationships and self.sample_relationships:
                self.predicate_context_extractor = FixedPredicateExtractor(fixed_predicates=self.fixed_relationships, predicate_types=self.sample_relationships)
            elif self.sample_relationships:
                self.predicate_context_extractor = FixedPredicateExtractor(predicate_types=self.sample_relationships)
            else:
                self.predicate_context_extractor = None
            self.create_emb = EmbeddingStore(inference_api_key=config.huggingface_token)
            if config.is_confined_search:
                self.llm_instance = Fixed_Entities_LLM(input_queue, llm_config)
            else :
                self.llm_instance = BERTLLM(input_queue, llm_config)
            self.rel_model_name = config.rel_model_name
            if config.openai_api_key:
                self.gpt_llm = OpenAI(api_key=config.openai_api_key)
            else:
                self.gpt_llm = OpenAI()
            self.function_registry = FunctionRegistry()
            
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to Initialize. {e}")
            raise Exception(f"Invalid {self.__class__.__name__} configuration. Unable to Initialize. {e}")
    
    def validate(self) -> bool:
        return isinstance(self.llm_instance, BERTLLM) or isinstance(self.llm_instance, Fixed_Entities_LLM)

    def process_messages(self, data: IngestedMessages):
        return super().process_messages(data)
    
    def process_images(self, data: IngestedImages):
        return super().process_messages(data)

    async def process_code(self, data: IngestedCode):
        return super().process_messages(data)

    @staticmethod
    def validate_ingested_tokens(data: IngestedTokens) -> bool:
        if data.is_error():
            
            return False

        return True

    @staticmethod
    def remove_items_from_tuples(data: List[Tuple[str, str, str]]) -> List[Tuple[str, str, str]]:
        result = []
        keys_to_remove = ['entity1_embedding', 'entity2_embedding', 'entity1_attnscore', 'entity2_attnscore', 'pair_attnscore']

        for tup in data:
            json_data = json.loads(tup[1])
            for key in keys_to_remove:
                json_data.pop(key, None)
            modified_json_str = json.dumps(json_data, ensure_ascii=False)
            modified_tuple = (tup[0], modified_json_str, tup[2])
            result.append(modified_tuple)

        return result
    
    async def process_triples(self, context, entity1, entity2, entity1_label, entity2_label):
        try:
            classify_entity_function = self.function_registry.get_classifyentity_function()
            predicate_info_function = self.function_registry.get_predicate_info_function()
            classify_entity_message = f"""Please analyze the provided context and two specified entities to identify the roles and types of these entities. These entities will be used to construct a semantic triple. A semantic triple is a structure used in semantic analysis and consists of three parts: a subject, a predicate, and an object. The subject is the main entity being discussed, the predicate is the action or relationship that connects the subject and object, and the object is the entity that is affected by or related to the subject. Use this information to the answer the user's query.
Context: {context} 
Entity 1: {entity1} and Entity 2: {entity2}
"""
            messages_classify_entity = [
                    {"role": "user", "content": classify_entity_message},
                    {"role": "user", "content": f"Query: Determine which entity is the subject and its respective type or category, also which entity is the object and its type or category e.g. location, time-period, rock, fossil , person, event, material, process etc."}
                    
                ]             
            classify_entity_response = self.generate_response(
                messages_classify_entity,
                classify_entity_function,
                "classify_entities"
            )
            subject_info = self.extract_subject_object_info(classify_entity_response)
            identify_predicate_message = f"Given the context, please identify the predicate between the subject '{subject_info['subject']}' and the object '{subject_info['object']}' and determine the predicate type e.g. causative, action, ownership, occurance etc."
            messages_identify_predicate = [
                                                {"role": "system", "content": identify_predicate_message},
                                                {"role": "user", "content": f"Context: {context}"}
                                            ]
            identify_predicate_response = self.generate_response(
                messages_identify_predicate,
                predicate_info_function,
                "predicate_info"
            )
            predicate_info = self.extract_predicate_info(identify_predicate_response)
            if not self.is_confined_search:
                return {
                    'subject_type': subject_info['subject_type'],
                    'subject': subject_info['subject'],
                    'object_type': subject_info['object_type'],
                    'object': subject_info['object'],
                    'predicate': predicate_info['predicate'],
                    'predicate_type': predicate_info['predicate_type']
                }
            else:
                if subject_info['subject'].lower() in entity1.lower() or entity1.lower() in subject_info['subject'].lower() or entity1.lower() == subject_info['subject'].lower():
                    return {
                    'subject_type': entity1_label,
                    'subject': subject_info['subject'],
                    'object_type': entity2_label,
                    'object': subject_info['object'],
                    'predicate': predicate_info['predicate'],
                    'predicate_type': predicate_info['predicate_type']
                }
                else:
                    return {
                    'subject_type': entity2_label,
                    'subject': subject_info['subject'],
                    'object_type': entity1_label,
                    'object': subject_info['object'],
                    'predicate': predicate_info['predicate'],
                    'predicate_type': predicate_info['predicate_type']
                }
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to process triples using GPT. {e}")

    # @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6))
    @retry(wait=wait_fixed(60), stop=stop_after_attempt(3))
    def completion_with_backoff(self, **kwargs):
        return self.gpt_llm.chat.completions.create(**kwargs)
    
    def generate_response(self, messages, functions, name):
        response = self.completion_with_backoff(
            model=self.rel_model_name,
            messages=messages,
            temperature=0,
            tools=functions,
            tool_choice="auto"
        )
        return response

    def extract_subject_object_info(self, response):
        function_call_arguments = json.loads(response.choices[0].message.tool_calls[0].function.arguments)
        if not function_call_arguments.get("subject") or not function_call_arguments.get("object"):
                raise ValueError("Missing 'subject',  or 'object' in llm output")
        return {
            'subject_type': function_call_arguments.get("subject_type", "Unlabeled"),
            'subject': function_call_arguments.get("subject"),
            'object_type': function_call_arguments.get("object_type", "Unlabeled"),
            'object': function_call_arguments.get("object")
        }

    def extract_predicate_info(self, response):
        function_call_arguments = json.loads(response.choices[0].message.tool_calls[0].function.arguments)
        if not function_call_arguments.get("predicate"):
                raise ValueError("Missing 'predicate' in llm output")
        return {
            'predicate': function_call_arguments.get("predicate"),
            'predicate_type': function_call_arguments.get("predicate_type", "Unlabeled")
        }

    def generate_output_tuple(self,result, context_json):
        context_data = json.loads(context_json)
        context = context_data.get("context", "")
        subject_type = result.get("subject_type", "Unlabeled")
        subject = result.get("subject", "")
        object_type = result.get("object_type", "Unlabeled")
        object = result.get("object", "")
        predicate = result.get("predicate", "")
        predicate_type = result.get("predicate_type", "Unlabled")
        output_tuple = (
            subject,
            f'{{"predicate": "{predicate}", "predicate_type": "{predicate_type}", "context": "{context}", "file": "{context_data.get("file_path", "")}", "subject_type": "{subject_type}", "object_type": "{object_type}"}}',
            object
        )

        return output_tuple
      
    async def process_tokens(self, data: IngestedTokens):
        try:
            if not GPTLLM.validate_ingested_tokens(data):
                    self.set_termination_event()                    
                    return 
            relationships = []
            result = await self.llm_instance.process_tokens(data)           
            if not result: return 
            else:
                filtered_triples, file = result
                modified_data = GPTLLM.remove_items_from_tuples(filtered_triples)
                for entity1, context_json, entity2 in modified_data:
                    context_data = json.loads(context_json)
                    context = context_data.get("context", "")
                    entity1_label = context_data.get("entity1_label", "")
                    entity2_label = context_data.get("entity2_label", "")
                    entity1_nn_chunk = context_data.get("entity1_nn_chunk","")
                    entity2_nn_chunk = context_data.get("entity2_nn_chunk","")
                    result = await self.process_triples(context, entity1_nn_chunk, entity2_nn_chunk, entity1_label, entity2_label)
                    if result:
                        output_tuple = self.generate_output_tuple(result, context_json)
                        relationships.append(output_tuple)
                if len(relationships) > 0:
                    embedding_triples = self.create_emb.generate_embeddings(relationships)
                    if self.sample_relationships:
                            embedding_triples = self.predicate_context_extractor.process_predicate_types(embedding_triples)
                    for triple in embedding_triples:
                        graph_json = json.dumps(TripleToJsonConverter.convert_graphjson(triple))
                        if graph_json:
                                current_state = EventState(EventType.Graph,1.0, graph_json, file)
                                await self.set_state(new_state=current_state)
                        vector_json = json.dumps(TripleToJsonConverter.convert_vectorjson(triple))
                        if vector_json:
                                current_state = EventState(EventType.Vector,1.0, vector_json, file)
                                await self.set_state(new_state=current_state)
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to extract predicates using GPT. {e}")
            raise Exception(f"An error occurred while extracting predicates using GPT: {e}")

    async def process_messages(self, data: IngestedMessages):
        raise NotImplementedError
